package envoy.client.ui.settings;

import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.*;

import envoy.client.Settings;
import envoy.client.event.EventBus;
import envoy.client.event.ThemeChangeEvent;
import envoy.client.ui.Theme;
import envoy.client.util.EnvoyLog;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>ThemeCustomizationPanel.java</strong><br>
 * Created: <strong>20 Dec 2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 */
public class ThemeCustomizationPanel extends SettingsPanel {

	private static final long serialVersionUID = -8697897390666456624L;

	private JPanel	colorsPanel				= new JPanel();
	private JButton	createNewThemeButton	= new JButton("Create New Theme");

	private String[]			themeArray	= Settings.getInstance().getThemes().keySet().toArray(new String[0]);
	private JComboBox<String>	themes		= new JComboBox<>(themeArray);
	private Theme				temporaryTheme, selectedTheme;

	private final Insets insets = new Insets(5, 5, 5, 5);

	private static final Logger logger = EnvoyLog.getLogger(ThemeCustomizationPanel.class.getSimpleName());

	/**
	 * Initializes a {@link ThemeCustomizationPanel} that enables the user to change
	 * the current {@link Theme} and create new themes as part of the
	 * {@link SettingsScreen}.
	 */
	public ThemeCustomizationPanel() {
		temporaryTheme = new Theme("temporaryTheme", Settings.getInstance().getThemes().get(Settings.getInstance().getCurrentTheme()));

		GridBagLayout gbl_themeLayout = new GridBagLayout();

		gbl_themeLayout.columnWidths	= new int[] { 1, 1 };
		gbl_themeLayout.rowHeights		= new int[] { 1, 1 };
		gbl_themeLayout.columnWeights	= new double[] { 1.0, 1.0 };
		gbl_themeLayout.rowWeights		= new double[] { 0.01, 1.0 };

		setLayout(gbl_themeLayout);

		themes.setSelectedItem(Settings.getInstance().getCurrentTheme());

		themes.addItemListener(new ItemListener() {

			@Override
			public void itemStateChanged(ItemEvent e) {
				String selectedValue = (String) themes.getSelectedItem();
				logger.log(Level.FINEST, selectedValue);
				selectedTheme = Settings.getInstance().getThemes().get(selectedValue);
			}
		});

		GridBagConstraints gbc_themes = new GridBagConstraints();
		gbc_themes.fill		= GridBagConstraints.HORIZONTAL;
		gbc_themes.gridx	= 0;
		gbc_themes.gridy	= 0;
		gbc_themes.anchor	= GridBagConstraints.NORTHWEST;
		gbc_themes.insets	= insets;

		add(themes, gbc_themes);

		colorsPanel.setLayout(new BoxLayout(colorsPanel, BoxLayout.Y_AXIS));
		colorsPanel.setAlignmentX(Component.LEFT_ALIGNMENT);

		Theme theme = Settings.getInstance().getThemes().get(Settings.getInstance().getCurrentTheme());
		buildCustomizeElements(theme);

		GridBagConstraints gbc_colorsPanel = new GridBagConstraints();
		gbc_colorsPanel.fill		= GridBagConstraints.HORIZONTAL;
		gbc_colorsPanel.gridx		= 0;
		gbc_colorsPanel.gridy		= 1;
		gbc_colorsPanel.gridwidth	= 2;
		gbc_colorsPanel.anchor		= GridBagConstraints.NORTHWEST;
		gbc_colorsPanel.insets		= insets;

		add(colorsPanel, gbc_colorsPanel);

		createNewThemeButton.setEnabled(false);
		createNewThemeButton.setBackground(theme.getInteractableBackgroundColor());
		createNewThemeButton.setForeground(theme.getInteractableForegroundColor());
		colorsPanel.setBackground(theme.getCellColor());

		createNewThemeButton.addActionListener((evt) -> {
			try {
				String name = JOptionPane.showInputDialog("Enter a name for the new theme");
				logger.log(Level.FINEST, name);
				Settings.getInstance().addNewThemeToMap(new Theme(name, temporaryTheme));
				themeArray							= Arrays.copyOf(themeArray, themeArray.length + 1);
				themeArray[themeArray.length - 1]	= Settings.getInstance().getThemes().get(name).getThemeName();

				temporaryTheme = new Theme("temporaryTheme", Settings.getInstance().getThemes().get(Settings.getInstance().getCurrentTheme()));

				createNewThemeButton.setEnabled(false);
				themes.addItem(themeArray[themeArray.length - 1]);

			} catch (Exception e) {
				logger.info("New theme couldn't be created! " + e);
				e.printStackTrace();
			}
		});

		GridBagConstraints gbc_createNewTheme = new GridBagConstraints();
		gbc_createNewTheme.gridx	= 0;
		gbc_createNewTheme.gridy	= 10;

		colorsPanel.add(createNewThemeButton, gbc_createNewTheme);

		// Apply current theme
		applyTheme(theme);

		// Respond to theme changes
		EventBus.getInstance().register(ThemeChangeEvent.class, (evt) -> applyTheme(((ThemeChangeEvent) evt).get()));
	}

	@Override
	public ActionListener getOkButtonAction() {
		return (evt) -> {
			Settings.getInstance().setCurrentTheme(selectedTheme.getThemeName());
			logger.log(Level.FINER, "Setting theme: " + selectedTheme.getThemeName());

			final Theme currentTheme = Settings.getInstance().getThemes().get(Settings.getInstance().getCurrentTheme());
			updateColorVariables(currentTheme);
			EventBus.getInstance().dispatch(new ThemeChangeEvent(currentTheme));

			revalidate();
			repaint();
		};
	}

	private void applyTheme(Theme theme) {
		// themeContent
		setForeground(theme.getUserNameColor());
		setBackground(theme.getCellColor());

		// themes
		themes.setBackground(theme.getBackgroundColor());
		themes.setForeground(getInvertedColor(theme.getBackgroundColor()));

		createNewThemeButton.setBackground(theme.getInteractableBackgroundColor());
		createNewThemeButton.setForeground(theme.getInteractableForegroundColor());
		colorsPanel.setBackground(theme.getCellColor());
	}

	private void updateColorVariables(Theme theme) {
		temporaryTheme = new Theme("temporaryTheme", theme);

		colorsPanel.removeAll();

		buildCustomizeElements(theme);

		GridBagConstraints gbc_createNewTheme = new GridBagConstraints();
		gbc_createNewTheme.gridx	= 0;
		gbc_createNewTheme.gridy	= 10;

		colorsPanel.add(createNewThemeButton, gbc_createNewTheme);
	}

	private void buildCustomizeElements(Theme theme) {
		buildCustomizeElement(theme, theme.getBackgroundColor(), "Background", "backgroundColor");
		buildCustomizeElement(theme, theme.getCellColor(), "Cells", "cellColor");
		buildCustomizeElement(theme, theme.getInteractableForegroundColor(), "Interactable Foreground", "interactableForegroundColor");
		buildCustomizeElement(theme, theme.getInteractableBackgroundColor(), "Interactable Background", "interactableBackgroundColor");
		buildCustomizeElement(theme, theme.getMessageColorChat(), "Messages Chat", "messageColorChat");
		buildCustomizeElement(theme, theme.getDateColorChat(), "Date Chat", "dateColorCat");
		buildCustomizeElement(theme, theme.getSelectionColor(), "Selection", "selectionColor");
		buildCustomizeElement(theme, theme.getTypingMessageColor(), "Typing Message", "typingMessageColor");
		buildCustomizeElement(theme, theme.getUserNameColor(), "User Names", "userNameColor");
	}

	private void buildCustomizeElement(Theme theme, Color color, String name, String colorName) {
		JPanel		panel		= new JPanel();
		JButton		button		= new JButton();
		JTextPane	textPane	= new JTextPane();

		textPane.setFont(new Font("Arial", Font.PLAIN, 14));
		textPane.setBackground(theme.getBackgroundColor());
		textPane.setForeground(getInvertedColor(theme.getBackgroundColor()));
		textPane.setText(name);
		textPane.setEditable(false);

		button.setBackground(color);
		button.setPreferredSize(new Dimension(25, 25));

		button.addActionListener((evt) -> {
			try {
				Color newColor = JColorChooser.showDialog(null, "Choose a color", color);
				if (newColor.getRGB() != color.getRGB()) {
					logger.log(Level.FINEST, "New Color: " + String.valueOf(color.getRGB()));
					// TODO: When Theme changed in same settings screen, color variable doesn't
					// update
					temporaryTheme.setColor(colorName, newColor);
					createNewThemeButton.setEnabled(true);
				}
				button.setBackground(newColor);

			} catch (Exception e) {
				logger.info("An error occured while opening Color Chooser: " + e);
				e.printStackTrace();
			}
		});

		panel.add(textPane);
		panel.add(button);
		panel.setBackground(theme.getCellColor());
		panel.setAlignmentX(Component.LEFT_ALIGNMENT);

		colorsPanel.add(panel);
	}

	private Color getInvertedColor(Color color) { return new Color(255 - color.getRed(), 255 - color.getGreen(), 255 - color.getBlue()); }
}
